from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from functools import partial
from typing import Optional, Tuple, List

import js
import wwwpy.remote.component as wpc
from js import document, console, Event, HTMLElement, window
from pyodide.ffi import create_proxy
from wwwpy.common import state, property_monitor, modlib
from wwwpy.common.designer import element_library
from wwwpy.common.designer.code_edit import add_component, ElementDef
from wwwpy.common.designer.element_path import ElementPath
from wwwpy.common.designer.html_edit import Position
from wwwpy.remote import dict_to_js, set_timeout
from wwwpy.remote.designer import element_path
from wwwpy.remote.designer.drop_zone import DropZone
from wwwpy.remote.designer.global_interceptor import GlobalInterceptor, InterceptorEvent
from wwwpy.remote.designer.ui.draggable_component import DraggableComponent
from wwwpy.server import rpc

from wwwpy.remote.designer.helpers import _element_lbl, _help_button
from wwwpy.remote.designer.ui.property_editor import PropertyEditor


@dataclass
class ToolboxState:
    geometry: Tuple[int, int, int, int] = field(default=(100, 100, 300, 250))
    toolbox_search: str = ''
    selected_element_path: Optional[ElementPath] = None


@dataclass
class MenuMeta:
    label: str
    html: str
    always_visible: bool = False
    p_element: js.HTMLElement = None


def menu(label, always_visible=False):
    def wrapped(fn):
        fn.label = label
        fn.meta = MenuMeta(label, label, always_visible)
        return fn

    return wrapped


class ToolboxComponent(wpc.Component, tag_name='wwwpy-toolbox'):
    body: HTMLElement = wpc.element()
    inputSearch: js.HTMLInputElement = wpc.element()
    dragComp1: DraggableComponent = wpc.element()
    property_editor: PropertyEditor = wpc.element()
    _select_element_btn: js.HTMLElement = wpc.element()
    _select_clear_btn: js.HTMLElement = wpc.element()
    components_marker = '-components-'

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
<style>
.two-column-layout {
  display: flex;
  flex-wrap: wrap;
}

.two-column-layout p {
  width: calc(50% - 10px); /* 50% width minus half of the gap */
  margin: 0 20px 10px 0; /* Right margin creates gap between columns */
}

.two-column-layout p:nth-child(even) {
  margin-right: 0; /* Removes right margin for every even child */
}
 .help-icon {
            width: 18px;
            height: 18px;
            vertical-align: middle;
            margin-left: 5px;
        }
</style>     
<svg xmlns="http://www.w3.org/2000/svg" style="display: none;">
    <symbol id="help-icon" viewBox="0 0 18 18">
        <circle cx="9" cy="9" r="8" fill="#007bff"/>
        <text x="9" y="9" fill="white" text-anchor="middle" dominant-baseline="central" font-weight="bold" font-size="12">?</text>
    </symbol>
</svg>   
<wwwpy-draggable-component data-name='dragComp1' style="height: 300px">
    <span slot='title'>wwwpy</span>     
        <button data-name="_select_element_btn">Select element...</button>   
        <button data-name="_select_clear_btn">Clear selection</button>   
        <wwwpy-property-editor data-name="property_editor"></wwwpy-property-editor>        
        <p><input data-name='inputSearch' type='search' placeholder='type to filter...'></p>
    <div data-name='body' class='two-column-layout'>
    </div>   
</wwwpy-draggable-component>         
"""
        self._manage_toolbox_state()

        attrs = [v for k, v in vars(self.__class__).items() if hasattr(v, 'label')]
        self._all_items: List[MenuMeta] = []

        def add_p(menu_meta: MenuMeta, callback):  # callback can be async
            self._all_items.append(menu_meta)
            p = document.createElement('p')
            menu_meta.p_element = p
            p.innerHTML = menu_meta.html
            p.addEventListener('click', create_proxy(callback))
            # p.menu_meta = menu_meta

        def add_comp(element_def: ElementDef):

            def _on_pointed(drop_zone: DropZone | None):
                console.log(f'pointed dropzone: {drop_zone}')
                msg = (f'Insert new {element_def.tag_name}')
                if drop_zone:
                    pos = 'before' if drop_zone.position == Position.beforebegin else 'after'
                    # msg += f' at {drop_zone.position.name} of {drop_zone.element.tagName}'
                    msg += f' {pos} {drop_zone.element.tagName}'
                else:
                    msg += ' ... select a dropzone.'
                self.property_editor.message1div.innerHTML = msg

            async def _start_drop_for_comp(event):
                _on_pointed(None)
                res = await _drop_zone_start_selection_async(_on_pointed)
                if res:
                    await self._process_dropzone(res, element_def)
                else:
                    self.property_editor.message1div.innerHTML = 'canceled'

            element_html = f'{element_def.tag_name} {_help_button(element_def)}'
            add_p(MenuMeta(element_def.tag_name, element_html), _start_drop_for_comp)

        for member in attrs:
            if member.meta.label == self.components_marker:
                [add_comp(ele_def) for ele_def in element_library.element_library().elements]
            else:
                add_p(member.meta, partial(member, self))
        self._update_toolbox_elements()

    async def _process_dropzone(self, drop_zone: DropZone, element_def: ElementDef):
        el_path = element_path.element_path(drop_zone.element)

        if not el_path:
            window.alert(f'No component found for dropzone!')
            return

        console.log(f'element_path={el_path}')
        file = modlib._find_module_path(el_path.class_module)
        old_source = file.read_text()

        add_result = add_component(old_source, el_path.class_name, element_def, el_path.path, drop_zone.position)

        if add_result:
            new_element_path = ElementPath(el_path.class_module, el_path.class_name, add_result.node_path)
            self._toolbox_state.selected_element_path = new_element_path
            await rpc.write_module_file(el_path.class_module, add_result.html)

    def _manage_toolbox_state(self):
        self._state_manager = state.State(state.JsStorage(), 'wwwpy.toolbox.state')
        restore = self._state_manager.restore(ToolboxState)
        self._toolbox_state: ToolboxState = restore.instance if restore.instance else ToolboxState()
        ts = self._toolbox_state

        def on_changes(*args):
            self._state_manager.save(self._toolbox_state)

        property_monitor.monitor_changes(self._toolbox_state, on_changes)

        self.inputSearch.value = ts.toolbox_search
        self.dragComp1.set_geometry(self._toolbox_state.geometry)

        def on_toolbar_geometry_change():
            if self.dragComp1.acceptable_geometry():
                self._toolbox_state.geometry = self.dragComp1.geometry()

        self.dragComp1.geometry_change_listeners.append(on_toolbar_geometry_change)
        self._restore_selected_element_path()

    def inputSearch__input(self, e: Event):
        self._toolbox_state.toolbox_search = self.inputSearch.value
        self._update_toolbox_elements()

    def _update_toolbox_elements(self):
        self.body.innerHTML = ''
        for meta in self._all_items:
            p = meta.p_element
            if meta.always_visible or self.inputSearch.value.lower() in meta.label.lower():
                self.body.appendChild(p)

    # @menu('Reload')
    # def _hot_reload(self, e: Event):
    #     _reload()
    async def _select_element_btn__click(self, e: Event):
        await self._select_component(e)

    async def _select_clear_btn__click(self, e: Event):
        self._toolbox_state.selected_element_path = None
        self._restore_selected_element_path()

    # @menu('Select component...', always_visible=True)
    async def _select_component(self, e: Event):
        no_comp = 'Click an element...'
        self.property_editor.message1div.innerHTML = no_comp

        def _on_pointed(drop_zone: DropZone | None):
            msg = no_comp
            if drop_zone:
                msg = 'Click to select ' + _element_lbl(drop_zone.element)
            self.property_editor.message1div.innerHTML = msg
            console.log(f'pointed dropzone: {drop_zone}')

        res = await _drop_zone_start_selection_async(_on_pointed, whole=True)
        self._toolbox_state.selected_element_path = element_path.element_path(res.element) if res else None
        self._restore_selected_element_path()

    @menu(components_marker)
    def _drop_zone_start(self, e: Event):
        assert False, 'Just a placeholder'

    @menu('handle global click')
    def _handle_global_click(self, e: Event):
        # add input
        def global_click(event: Event):
            event.preventDefault()
            event.stopImmediatePropagation()
            event.stopPropagation()
            console.log('global click', event.element)
            if self._global_click:
                document.removeEventListener('click', self._global_click, True)
                self._global_click = None

        self._global_click = create_proxy(global_click)
        document.addEventListener('click', self._global_click, True)

    def _restore_selected_element_path(self):
        element_path = self._toolbox_state.selected_element_path
        if element_path:
            console.log(f'restoring selected element path: {element_path}')
            if not element_path.valid():
                element_path = None

        self.property_editor.selected_element_path = element_path
        if element_path:
            self._select_clear_btn.removeAttribute('disabled')
        else:
            self._select_clear_btn.setAttribute('disabled', '')


def is_inside_toolbar(element: HTMLElement | None):
    if not element:
        return False

    orig = element
    # loop until the root element and see if it is the toolbar
    res = False
    while element:
        if element.tagName.lower() == ToolboxComponent.component_metadata.tag_name:
            res = True
            break
        element = element.parentElement

    return res


def _default_drop_zone_accept(drop_zone: DropZone):
    name = drop_zone.element.tagName.lower()
    accept = not (name == 'body' or name == 'html' or name == ToolboxComponent.component_metadata.tag_name)
    return accept


async def _drop_zone_start_selection_async(on_pointed, whole=False) -> Optional[DropZone]:
    from wwwpy.remote.designer.drop_zone import drop_zone_selector

    event = asyncio.Event()
    result = []

    def intercept_ended(ev: InterceptorEvent):
        selected = drop_zone_selector.stop()
        if is_inside_toolbar(ev.target):
            ev.uninstall()
            event.set()
            return
        ev.preventAndStop()
        if selected:
            ev.uninstall()
            console.log(
                f'selection accepted position {selected.position.name} target: ', selected.element,
                'parent: ', selected.element.parentElement,
                'event: ', ev.event,
                'composedPath: ', ev.event.composedPath(),
            )
            result.append(selected)
            event.set()

    GlobalInterceptor(intercept_ended, 'pointerdown').install()

    click_inter = GlobalInterceptor(lambda ev: ev.preventAndStop(), 'click')
    click_inter.install()
    drop_zone_selector.start_selector(on_pointed, _default_drop_zone_accept, whole=whole)
    await event.wait()
    set_timeout(lambda: click_inter.uninstall(), 500)

    if len(result) == 0:
        return None

    return result[0]



from __future__ import annotations

import js
from js import Array, Element, document

from wwwpy.common.designer.element_path import ElementPath
from wwwpy.common.designer.html_locator import Node, NodePath


def _fqn(obj):
    return f"{obj.__class__.__module__}.{obj.__class__.__name__}"


def element_path(element: Element) -> ElementPath | None:
    """Returns an instance of  """

    path = []
    while element:
        if hasattr(element, "_py"):
            # todo this should be moved to component.py in a function like component_from_element
            # or get_underlying_component
            component = element._py
            if hasattr(component, "unwrap"):
                component = component.unwrap()
            clazz = component.__class__
            return ElementPath(clazz.__module__, clazz.__name__, path)
        if element == document.body:
            return None

        parent = element.parentNode
        child_index = Array.prototype.indexOf.call(parent.children, element) if parent else -1
        attributes = {attr.name: attr.value for attr in element.attributes}
        path.insert(0, Node(element.tagName.lower(), child_index, attributes))
        element = parent

    return None


def element_to_node_path(element: HTMLElement) -> NodePath:
    """
    Get the path from the root to the target.
    """

    path = []
    while element:
        if element == document.body:
            return path
        parent = element.parentNode
        child_index = Array.prototype.indexOf.call(parent.children, element) if parent else -1
        attributes = {attr.name: attr.value for attr in element.attributes}
        path.insert(0, Node(element.tagName.lower(), child_index, attributes))
        element = parent

    return path



from __future__ import annotations

from typing import Callable, Union, Awaitable


def dict_to_js(o):
    import js
    import pyodide
    return pyodide.ffi.to_js(o, dict_converter=js.Object.fromEntries)


async def micropip_install(package):
    from js import pyodide
    await pyodide.loadPackage('micropip')
    import micropip
    await micropip.install([package])


def set_timeout(callback: Callable[[], Union[None, Awaitable[None]]], timeout_millis: int | None = 0):
    from pyodide.ffi import create_once_callable
    from js import window
    window.setTimeout(create_once_callable(callback), timeout_millis)



from typing import List

from wwwpy.common.designer.element_library import ElementDef, Help, EventDef


def _standard_elements_def() -> List[ElementDef]:
    res = [
        ElementDef(
            'button', 'js.HTMLButtonElement',
            help=Help('A clickable button.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/button'),
            events=[EventDef('click', Help('The button was clicked.',
                                           'https://developer.mozilla.org/en-US/docs/Web/API/Element/click_event'))],
        ),
        ElementDef(
            'input', 'js.HTMLInputElement',
            help=Help('A field for entering text.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input'),
            events=[
                EventDef('input', Help('The input event fires when the value of the element has been changed '
                                       'as a direct result of a user action',
                                       'https://developer.mozilla.org/en-US/docs/Web/API/Element/input_event')),
                EventDef('change')
            ]
        ),
        ElementDef(
            'div', 'js.HTMLDivElement',
            help=Help('A generic container element.',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/div')
        ),

    ]
    for r in res:
        r.gen_html = _generateHtml
    return res


def _generateHtml(element_def: ElementDef, name: str) -> str:
    tag_name = element_def.tag_name

    def _def(placeHolder=False, add=''):
        def inner():
            pl = '' if not placeHolder else f' placeholder="{name}"'
            add1 = '' if not add else f' {add}'
            return f'<{tag_name} data-name="{name}"{pl}{add1}>{name}</{tag_name}>'

        return inner

    func = {
        'button': _def(),
    }
    gen_html = func.get(tag_name, None)
    html = '\n' + gen_html() if gen_html else '' + ElementDef.default_gen_html(element_def, name)
    return html


# [
#     _comp('input', 'js.HTMLInputElement', '<input type="text" data-name="#name#" value="#name#">'),
#     _comp('button', 'js.HTMLButtonElement', '<button data-name="#name#">#name#</button>'),
#     _comp('textarea', 'js.HTMLTextAreaElement', '<textarea data-name="#name#">#name#</textarea>'),
#     _comp('select', 'js.HTMLSelectElement',
#           '<select data-name="#name#"><option>#name#</option></select>'),
#     _comp('div', 'js.HTMLDivElement', '<div data-name="#name#">#name#</div>'),
#     _comp('p', 'js.HTMLParagraphElement', '<p data-name="#name#">#name#</p>'),
#     _comp('h1', 'js.HTMLHeadingElement', '<h1 data-name="#name#">#name#</h1>'),
#     _comp('h2', 'js.HTMLHeadingElement', '<h2 data-name="#name#">#name#</h2>'),
#     _comp('h3', 'js.HTMLHeadingElement', '<h3 data-name="#name#">#name#</h3>'),
#     _comp('a', 'js.HTMLAnchorElement', '<a href="#" data-name="#name#">#name#</a>'),
#     _comp('img', 'js.HTMLImageElement', '<img src="#" alt="#name#" data-name="#name#">'),
#     _comp('ul', 'js.HTMLUListElement', '<ul data-name="#name#"><li>#name#</li></ul>'),
#     _comp('ol', 'js.HTMLOListElement', '<ol data-name="#name#"><li>#name#</li></ol>'),
#     _comp('li', 'js.HTMLLIElement', '<li data-name="#name#">#name#</li>'),
#     _comp('table', 'js.HTMLTableElement', '<table data-name="#name#"><tr><td>#name#</td></tr></table>'),
#     _comp('form', 'js.HTMLFormElement', '<form data-name="#name#">#name#</form>'),
#     _comp('label', 'js.HTMLLabelElement', '<label data-name="#name#">#name#</label>'),
#
# ]


from datetime import datetime
from pathlib import Path
from typing import Callable

from watchdog.events import FileSystemEventHandler, FileSystemEvent, DirModifiedEvent
from watchdog.observers import Observer

from wwwpy.common.throttler import Event
from wwwpy.server.throttler_thread import EventThrottlerThread

_default_folders = {'remote', 'common'}


class ChangeHandler(FileSystemEventHandler):

    def __init__(self, path: Path, callback: Callable[[FileSystemEvent], None], folders: set[str] = _default_folders):
        """path need to exist, otherwise the observer will throw an exception. Path should be the root of wwwpy working dir.
        folders is a set of folder names (children of path) that are monitored for changes.
        """
        self._path = path
        self._callback = callback
        self._folders = map(lambda f: path / f, folders)
        self._observer = Observer()

        def _emit(event: Event):
            self._callback(event.item)

        self._throttler = EventThrottlerThread(100, _emit, datetime.utcnow)
        super().__init__()

    def on_any_event(self, event: FileSystemEvent) -> None:
        if Path(event.src_path) == self._path:
            return
        if isinstance(event, DirModifiedEvent):
            return
        src_path = Path(event.src_path)
        rel = src_path.relative_to(self._path)
        if rel.parts[0] not in _default_folders:
            return
        if event.event_type == 'closed' or event.event_type == 'opened' or event.event_type == 'created':
            return

        key = str(event.src_path)
        if key.endswith('.py~'):
            return
        self._throttler.new_event(Event(key, event))

    def watch_directory(self):
        self._observer.schedule(self, str(self._path), recursive=True)
        self._observer.start()

    def stop_join(self):
        self._observer.stop()
        self._observer.join()



from __future__ import annotations

from dataclasses import dataclass

import libcst as cst

from wwwpy.common.designer import code_info, html_parser, html_locator
from wwwpy.common.designer.code_info import Attribute
from wwwpy.common.designer.code_strings import html_string_edit
from wwwpy.common.designer.element_library import ElementDef
from wwwpy.common.designer.html_edit import Position, html_add
from wwwpy.common.designer.html_locator import NodePath


def add_property(source_code: str, class_name: str, attr_info: Attribute):
    source_code_imp = ensure_imports(source_code)
    module = cst.parse_module(source_code_imp)
    transformer = _AddFieldToClassTransformer(class_name, attr_info)
    modified_tree = module.visit(transformer)

    return modified_tree.code


@dataclass
class AddResult:
    html: str
    node_path: NodePath


def add_component(source_code: str, class_name: str, comp_def: ElementDef, node_path: NodePath,
                  position: Position) -> AddResult | None:
    source_code = ensure_imports(source_code)
    class_info = code_info.class_info(source_code, class_name)
    if class_info is None:
        print(f'Class {class_name} not found inside source ```{source_code}```')
        return None

    attr_name = class_info.next_attribute_name(comp_def.tag_name)
    named_html = comp_def.new_html(attr_name)

    source1 = add_property(source_code, class_name, Attribute(attr_name, comp_def.python_type, 'wpc.element()'))

    def manipulate_html(html):
        add = html_add(html, named_html, node_path, position)
        return add

    source2 = html_string_edit(source1, class_name, manipulate_html)
    new_tree = html_parser.html_to_tree(source2)
    displacement = 0 if position == Position.beforebegin else 1
    indexes = [n.child_index for n in node_path[0:-1]] + [node_path[-1].child_index + displacement]
    new_node_path = html_locator.tree_to_path(new_tree, indexes)
    return AddResult(source2, new_node_path)


class _AddFieldToClassTransformer(cst.CSTTransformer):
    def __init__(self, class_name, new_field: Attribute):
        super().__init__()
        self.class_name = class_name
        self.new_field = new_field

    def leave_ClassDef(self, original_node, updated_node):
        if original_node.name.value != self.class_name:
            return original_node
        # Check if the type is a composite name (contains a dot)
        if '.' in self.new_field.type:
            base_name, attr_name = self.new_field.type.rsplit('.', 1)
            annotation = cst.Annotation(cst.Attribute(value=cst.Name(base_name), attr=cst.Name(attr_name)))
        else:
            annotation = cst.Annotation(cst.Name(self.new_field.type))

        new_field_node = cst.SimpleStatementLine([
            cst.AnnAssign(
                target=cst.Name(self.new_field.name),
                annotation=annotation,
                value=None if self.new_field.default is None else cst.parse_expression(self.new_field.default)
            )
        ])

        # Find the position to insert the new attribute
        last_assign_index = 0
        for i, item in enumerate(updated_node.body.body):
            if isinstance(item, cst.SimpleStatementLine) and isinstance(item.body[0], cst.AnnAssign):
                last_assign_index = i + 1

        new_body = list(updated_node.body.body)
        new_body.insert(last_assign_index, new_field_node)

        return updated_node.with_changes(body=updated_node.body.with_changes(body=new_body))


def add_method(source_code: str, class_name: str, method_name: str, method_args: str,
               instructions: str = 'pass') -> str:
    module = cst.parse_module(source_code)
    transformer = _AddMethodToClassTransformer(True, class_name, method_name, 'self, ' + method_args, instructions)
    modified_tree = module.visit(transformer)
    return modified_tree.code


class _AddMethodToClassTransformer(cst.CSTTransformer):
    def __init__(self, async_def: bool, class_name: str, method_name: str, method_args: str, instructions: str):
        super().__init__()
        self.async_def = async_def
        self.class_name = class_name
        self.method_name = method_name
        self.method_args = method_args
        self.instructions = instructions

    def leave_ClassDef(self, original_node, updated_node):
        if original_node.name.value != self.class_name:
            return original_node

        parsed_instructions = cst.parse_module(self.instructions).body

        new_method_node = cst.FunctionDef(
            name=cst.Name(self.method_name),
            params=cst.Parameters(
                params=[cst.Param(name=cst.Name(arg.strip())) for arg in self.method_args.split(',') if arg.strip()]
            ),
            body=cst.IndentedBlock(body=parsed_instructions),
            asynchronous=cst.Asynchronous() if self.async_def else None
        )

        new_body = list(updated_node.body.body)
        new_body.append(cst.EmptyLine())
        new_body.append(new_method_node)
        new_body.append(cst.EmptyLine())

        return updated_node.with_changes(body=updated_node.body.with_changes(body=new_body))


def ensure_imports(source_code: str) -> str:
    required_imports = [
        'import wwwpy.remote.component as wpc',
        'import js'
    ]

    def _remove_comment_if_present(line) -> str:
        line = line.strip()
        if '#' in line:
            line = line[:line.index('#')]
        return line.strip()

    existing_imports = set(_remove_comment_if_present(line) for line in source_code.split('\n')
                           if line.strip().startswith('import'))

    for imp in required_imports:
        if imp not in existing_imports:
            source_code = imp + '\n' + source_code

    return source_code



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from queue import PriorityQueue
from threading import Lock
from typing import Any, Callable, Dict


class State(Enum):
    FIRST = 1
    THROTTLE = 2
    DELAYED = 3


@dataclass
class Event:
    key: Any = field()
    item: Any = field()


@dataclass(order=True)
class _EventState:
    action_time: datetime = field()
    event: Event = field(compare=False)
    state: State = field(compare=False, default=State.FIRST)


class EventThrottler:
    def __init__(
            self, timeout_millis: int,
            emit: Callable[[Event], None],
            time_provider: Callable[[], datetime]):
        """A wakeup call should wake up the thread. The thread should call process_queue to process the events.
        and then go to sleep accordingly to next_action_delta."""
        if time_provider is None:
            time_provider = datetime.utcnow
        self._time_provider = time_provider
        self._emit = emit
        self._timeout_millis = timeout_millis

        # The following two data structure needs to be intended as a single atomic datastructure
        # if an item appears/disappear from one it should appear/disappear from the other
        self._queue: PriorityQueue[_EventState] = PriorityQueue()
        self._states: Dict[Any, _EventState] = {}
        self._lock = Lock()

    def new_event(self, event: Event):
        with self._lock:
            prev_evt = self._states.get(event.key, None)
            if prev_evt is None:
                state = _EventState(self._time_provider(), event)  # add for immediate emit
                self._states[event.key] = state
                self._queue.put(state)
            else:
                assert prev_evt.event.key == event.key
                prev_evt.event = event
                prev_evt.state = State.DELAYED

            self.wakeup()

    def process_queue(self):
        def done():
            delta = self.next_action_delta()
            return delta is None or delta.total_seconds() > 0

        while not done():
            with self._lock:

                evt = self._queue.get()
                if evt.state == State.FIRST:
                    evt.state = State.THROTTLE
                    evt.action_time += timedelta(milliseconds=self._timeout_millis)
                    self._queue.put(evt)
                else:
                    del self._states[evt.event.key]
                    if evt.state == State.DELAYED:
                        pass  # we need to emit this
                    elif evt.state == State.THROTTLE:
                        evt = None  # we already emitted when it was in FIRST state

            if evt is not None:
                self._emit(evt.event)

    def next_action_delta(self) -> timedelta | None:
        if self._queue.empty():
            return None
        evt = self._queue.queue[0]
        diff = evt.action_time - self._time_provider()
        return diff

    def wakeup(self):
        raise NotImplementedError



import sys

from wwwpy.common import modlib


async def write_module_file(module: str, content: str):
    path = modlib._find_module_path(module)
    if not path:
        raise ValueError(f'Cannot find module {module}')
    print(f'writing={path.name}, content length={len(content)}')
    path.write_text(content)


def _fix_stacktrace(message: str):
    return message.replace('"/wwwpy_bundle/', '"')


async def on_error(message: str, source: str, lineno: int, colno: int, error: str):
    print(f'rpc.on_error')
    message = _fix_stacktrace(message)
    error = _fix_stacktrace(error)
    all_str = f'message==={message}\nsource==={source}\nlineno==={lineno}\ncolno==={colno}\nerror==={error}'
    print(all_str, file=sys.stderr)


async def on_unhandledrejection(message: str):
    message = _fix_stacktrace(message)
    print(message, file=sys.stderr)


async def print_module_line(module: str, message: str, lineno: int):
    #  File "remote/designer/toolbar.py", line 140, in on_changes
    path = modlib._find_module_path(module)
    if not path:
        raise ValueError(f'Cannot find module {module}')
    print(message)
    print(f' File "{path}", line {lineno}')


async def server_console_log(message: str):
    print(message)



from __future__ import annotations

import sys
from pathlib import Path

from wwwpy.bootstrap import bootstrap_routes
from wwwpy.common.rpc.custom_loader import CustomFinder
from wwwpy.resources import library_resources, from_directory, from_file
from wwwpy.webserver import Webserver
from wwwpy.webservers.available_webservers import available_webservers
from wwwpy.websocket import WebsocketPool


def start_default(port: int, directory: Path, dev_mode=False):
    webserver = available_webservers().new_instance()

    convention(directory, webserver, dev_mode=dev_mode)

    webserver.set_port(port).start_listen()


websocket_pool: WebsocketPool = None


def convention(directory: Path, webserver: Webserver = None, dev_mode=False):
    print(f'applying convention to working_dir: {directory}')
    if dev_mode:
        from wwwpy.common import quickstart
        quickstart._check_quickstart(directory)

    sys.path.insert(0, str(directory))
    sys.meta_path.insert(0, CustomFinder({'remote', 'remote.rpc', 'wwwpy.remote', 'wwwpy.remote.rpc'}))
    global websocket_pool
    websocket_pool = WebsocketPool('/wwwpy/ws')
    services = _configure_rpc_services('/wwwpy/rpc', ['wwwpy.server.rpc', 'server.rpc'])
    routes = [services.route, websocket_pool.http_route, *bootstrap_routes(
        resources=[
            library_resources(),
            services.remote_stub_resources(),
            from_directory(directory / 'remote', relative_to=directory),
            from_directory(directory / 'common', relative_to=directory),
            from_file(directory / 'common.py', relative_to=directory),  # remove .py support
            from_file(directory / 'remote.py', relative_to=directory),  # remove .py support
        ],
        # language=python
        python=f'from wwwpy.remote.browser_main import entry_point; await entry_point(dev_mode={dev_mode})'
    )]

    if dev_mode:
        from wwwpy.server.designer import dev_mode as dv
        dv._hotreload_remote(['common', 'remote'], websocket_pool)
        dv._hotreload_server(['common', 'server'])

    if webserver is not None:
        webserver.set_http_route(*routes)


from wwwpy.rpc import RpcRoute, Module


def _configure_rpc_services(route_path: str, modules: list[str]) -> RpcRoute:
    services = RpcRoute(route_path)

    for module_name in modules:
        try:
            services.add_module(module_name)
        except Exception as e:
            print(f'could not load rpc module `{module_name}`: {e}')
            return None

    return services



import shutil
from pathlib import Path


def _setup_quickstart(directory: Path):
    source = Path(__file__).parent / 'setup1'
    shutil.copytree(source, directory, dirs_exist_ok=True)


def _check_quickstart(directory: Path):
    if _do_quickstart(directory):
        print(f'empty directory, creating quickstart in {directory.absolute()}')
        _setup_quickstart(directory)
    else:
        (directory / 'remote').mkdir(exist_ok=True)


def _do_quickstart(directory):
    return all(item.name.startswith('.') for item in directory.iterdir())


import tempfile
from pathlib import Path
from typing import List, Protocol, Any

from wwwpy.common.filesystem.sync.event import Event


def new_tmp_path() -> Path:
    return Path(tempfile.mkdtemp(prefix='debounce-tmp-path-'))


def filesystemevents_print(events: List[Event], package:str=''):
    # for e in events:
    #     print(f'  {e}')
    def accept(e: Event) -> bool:
        bad = e.is_directory and e.event_type == 'modified'
        return not bad
    def to_str(e: Event) -> str:
        return e.src_path if e.dest_path == '' else f'{e.src_path} -> {e.dest_path}'
    joined = set(to_str(e) for e in events if accept(e))
    summary = ', '.join(joined)
    if package != '':
        package = f' for `{package}`'
    print(f'Hotreload event{package} count: {len(events)}. Changes summary: {summary}')


class Sync(Protocol):

    @staticmethod
    def sync_source(source: Path, events: List[Event]) -> List[Any]:
        """Could be called 'sync_produce'. This part has the following input:
- a reference to the root of the monitored source filesystem
- a list of filesystem change events


The output is:
- a list of filesystem changes that can be serialized.
- such list can be an aggregated list of changes (e.g., a file was created and then modified, the two events can be aggregated into a single event)

"""

    @staticmethod
    def sync_target(target: Path, changes: List[Any]) -> None:
        """Could be called 'sync_apply'. This part has the following input:
- a reference to the root of the target filesystem
- a list of filesystem aggregated change events

The output is:
- the target filesystem is updated to reflect the changes
"""

    @staticmethod
    def sync_init(source: Path) -> List[Any]:
        """This part has the following input:
- a reference to the root of the monitored source filesystem

The output is:
- a list of filesystem aggregated change events that reflect the current state of the source filesystem
"""



from __future__ import annotations

from typing import Tuple, List, Callable

import js
from js import document, console, ResizeObserver
from pyodide.ffi import create_proxy
from pyodide.ffi.wrappers import add_event_listener, remove_event_listener

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js


class DraggableComponent(wpc.Component, tag_name='wwwpy-draggable-component'):
    container_div: wpc.HTMLElement = wpc.element()
    draggable_component_div: wpc.HTMLElement = wpc.element()
    resize_handle: wpc.HTMLElement = wpc.element()
    client_x = 0
    client_y = 0
    css_border = 2  # 1px border on each side, so we need to subtract 2px from width and height
    geometry_change_listeners: List[Callable[[], None]] = []

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
<style>
.wwwpy-container_div {
  position: absolute;
  z-index: 100000;
  background-color: black;
  border: 1px solid #d3d3d3;
  text-align: center;
  resize: both;
  overflow: auto;
}

.wwwpy-draggable_component_div {
  padding: 10px;
  cursor: move;
  z-index: 1001;
  background-color: #2196F3;
  color: #fff;
}

</style>        
<div data-name="container_div" class='wwwpy-container_div'>
    <div  data-name="draggable_component_div" class='wwwpy-draggable_component_div' >
        <slot name='title' >slot=title</slot>
    </div>
    <slot>slot=default</div>    
</div> 
"""
        self.client_x = 0
        self.client_y = 0
        tb = self.container_div

        def tb_print(*args):
            console.log(f'offsets of container_div: {self.geometry()}')

        tb.print = tb_print

        def on_resize(entries, observer):
            tb.print()
            self._on_geometry_change()

        resize_observer = ResizeObserver.new(create_proxy(on_resize))
        resize_observer.observe(self.container_div)

    def _on_geometry_change(self):
        for listener in self.geometry_change_listeners:
            listener()

    def draggable_component_div__touchstart(self, e: js.TouchEvent):
        self._move_start(e)

    def draggable_component_div__mousedown(self, e: js.MouseEvent):
        self._move_start(e)

    def _move_start(self, e: js.MouseEvent | js.TouchEvent):
        e.preventDefault()
        self.client_x = clientX(e)  # e.clientX
        self.client_y = clientY(e)  # e.clientY
        add_event_listener(document, 'mousemove', self._move)
        add_event_listener(document, 'mouseup', self._move_stop)
        add_event_listener(document, 'touchmove', self._move)
        add_event_listener(document, 'touchend', self._move_stop)

    def _move(self, event: js.MouseEvent | js.TouchEvent):
        # event.preventDefault()
        x = clientX(event)
        y = clientY(event)
        delta_x = self.client_x - x
        delta_y = self.client_y - y
        self.client_x = x
        self.client_y = y

        new_left = self.container_div.offsetLeft - delta_x
        new_top = self.container_div.offsetTop - delta_y

        self.set_position(new_left, new_top)
        self._on_geometry_change()

    def _move_stop(self, event):
        remove_event_listener(document, 'mousemove', self._move)
        remove_event_listener(document, 'mouseup', self._move_stop)
        remove_event_listener(document, 'touchmove', self._move)
        remove_event_listener(document, 'touchend', self._move_stop)

    def geometry(self) -> Tuple[int, int, int, int]:
        t = self.container_div
        return t.offsetTop, t.offsetLeft, (t.offsetWidth - self.css_border), (t.offsetHeight - self.css_border)

    def set_geometry(self, geometry_tuple):
        top, left, width, height = geometry_tuple
        self.set_position(left, top)
        self.set_size(f"{height}px", f"{width}px")

    def set_position(self, left, top):
        self.container_div.style.top = f"{top}px"
        self.container_div.style.left = f"{left}px"

    def set_size(self, h, w):
        self.container_div.style.width = w
        self.container_div.style.height = h

    def acceptable_geometry(self) -> bool:
        top, left, width, height = self.geometry()
        return width > 100 and height > 100


def clientX(event: js.MouseEvent | js.TouchEvent):
    #   var top = e.clientY || e.targetTouches[0].pageY;
    # return event.clientX if hasattr(event, 'clientX') else event.targetTouches[0].clientX
    if hasattr(event, 'targetTouches'):
        return list(event.targetTouches)[0].clientX
    return event.clientX


def clientY(event: js.MouseEvent | js.TouchEvent):
    return event.clientY if hasattr(event, 'clientY') else list(event.targetTouches)[0].clientY



from typing import Protocol, List, Any


class FileChangedListener(Protocol):
    def __call__(self, package: str, events: List[Any]):
        ...


_file_changed_listeners: List[FileChangedListener] = []


class BrowserRpc:

    def package_file_changed_sync(self, package: str, events: List[Any]):
        for listener in _file_changed_listeners:
            listener(package, events)


def file_changed_listeners_add(listener: FileChangedListener):
    _file_changed_listeners.append(listener)



from threading import Thread, Event
from typing import Callable

from wwwpy.common.throttler import EventThrottler


class EventThrottlerThread(EventThrottler):
    def __init__(self, max_delay_millis, emit: Callable[[Event], None], time_provider):
        self._thread = Thread(daemon=True, target=self._thread_loop, name='EventThrottlerThread')
        self._event = Event()
        super().__init__(max_delay_millis, emit, time_provider)
        self._thread.start()

    def wakeup(self):
        self._event.set()

    def _thread_loop(self):
        while self._thread is not None:
            delta = self.next_action_delta()
            timeout_millis = None if delta is None else delta.total_seconds() * 1000
            self._event.wait(timeout_millis)
            self._event.clear()
            self.process_queue()

    def stop_join(self):
        t = self._thread
        self._thread = None
        self.wakeup()
        t.join()



from __future__ import annotations

import sys
from typing import List, Any

from wwwpy.common import modlib
from wwwpy.remote import rpc as rpc
from js import console

_wwwpy_dev_mode = 'wwwpy_dev_mode'


async def activate():
    setattr(sys, _wwwpy_dev_mode, True)
    from wwwpy.remote import micropip_install
    from wwwpy.common import designer

    from wwwpy.common.filesystem.sync import sync_delta2
    from wwwpy.common.filesystem.sync import Sync
    sync_impl: Sync = sync_delta2

    for package in designer.pypi_packages:
        await micropip_install(package)

    def file_changed(package_name: str, events: List[Any]):
        directory = modlib._find_package_directory(package_name)
        sync_impl.sync_target(directory, events)

        from wwwpy.remote.browser_main import _reload
        _reload()

    rpc.file_changed_listeners_add(file_changed)


def is_active() -> bool:
    return hasattr(sys, _wwwpy_dev_mode)



from __future__ import annotations

from datetime import timedelta
from functools import partial
from typing import List, Callable

import wwwpy.remote.rpc as rpc
from wwwpy.common import modlib
from wwwpy.common.filesystem import sync
from wwwpy.common.filesystem.sync import Sync
from wwwpy.common.filesystem.sync import filesystemevents_print
from wwwpy.common.filesystem.sync import sync_delta2
from wwwpy.server.filesystem_sync.watchdog_debouncer import WatchdogDebouncer
from wwwpy.websocket import WebsocketPool

sync_impl: Sync = sync_delta2


def _watch_filesystem_change_for_remote(package: str, websocket_pool: WebsocketPool):
    directory = modlib._find_package_directory(package)
    if not directory:
        return

    def on_sync_events(events: List[sync.Event]):
        try:
            filesystemevents_print(events, package)
            payload = sync_impl.sync_source(directory, events)
            for client in websocket_pool.clients:
                remote_rpc = client.rpc(rpc.BrowserRpc)
                remote_rpc.package_file_changed_sync(package, payload)
        except:
            # we could send a sync_init
            import traceback
            print(f'on_sync_events 1 {traceback.format_exc()}')

    handler = WatchdogDebouncer(directory, timedelta(milliseconds=100), on_sync_events)
    handler.watch_directory()


def _hotreload_remote(remote_packages: list[str], websocket_pool: WebsocketPool, ):
    for package in remote_packages:
        _watch_filesystem_change_for_remote(package, websocket_pool)


def _watch_filesystem_change_for_server(package: str, callback: Callable[[str, List[sync.Event]], None]):
    directory = modlib._find_package_directory(package)
    if not directory:
        return

    def on_sync_events(events: List[sync.Event]):
        try:
            filesystemevents_print(events)
            callback(package, events)
        except:
            import traceback
            print(f'_watch_filesystem_change_for_server {traceback.format_exc()}')

    handler = WatchdogDebouncer(directory, timedelta(milliseconds=100), on_sync_events)
    handler.watch_directory()


def _hotreload_server(hotreload_packages: list[str]):
    def on_change(package: str, events: List[sync.Event]):

        for p in hotreload_packages:
            directory = modlib._find_package_directory(p)
            if directory:
                try:
                    import wwwpy.common.reloader as reloader
                    reloader.unload_path(str(directory))
                except:
                    # we could send a sync_init
                    import traceback
                    print(f'_hotreload_server {traceback.format_exc()}')

    for package in hotreload_packages:
        _watch_filesystem_change_for_server(package, on_change)



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Dict, List, Tuple, Optional

from wwwpy.common.designer.html_parser import html_to_tree, CstNode


@dataclass()
class Node:
    tag_name: str
    """The tag name in lowercase."""
    child_index: int
    """This is the index in the list of children of the parent node.
    It is -1 if the node has no parent.
    """
    attributes: Dict[str, Optional[str]]
    """The HTML attributes of the node."""

    def __post_init__(self):
        assert self.tag_name == self.tag_name.lower()


NodePath = List[Node]
"""This is the path from the root to a node in the DOM tree."""


def node_path_serialize(path: NodePath) -> str:
    return json.dumps([node.__dict__ for node in path])


def node_path_deserialize(serialized: str) -> NodePath:
    node_dicts = json.loads(serialized)
    return [Node(**node_dict) for node_dict in node_dicts]


def locate_node(html: str, path: NodePath) -> CstNode | None:
    cst_tree = html_to_tree(html)

    def find_node(nodes: List[CstNode], path: NodePath, depth: int) -> CstNode | None:
        if depth >= len(path):
            return None

        target_node = path[depth]
        if target_node.child_index < 0 or target_node.child_index >= len(nodes):
            return None
        node = nodes[target_node.child_index]
        if depth == len(path) - 1:
            return node
        return find_node(node.children, path, depth + 1)

    target_node = find_node(cst_tree, path, 0)

    return target_node


def locate(html: str, path: NodePath) -> Tuple[int, int] | None:
    """This function locates the position of the node specified by the path in the HTML string.
    The position is represented by the start and end indices of the node in the HTML string.
    """

    target_node = locate_node(html, path)
    if target_node:
        return target_node.span
    return None


def tree_to_path(tree: List[CstNode], indexed_path: list[int]) -> NodePath:
    """This function converts a tree of CstNode objects to a NodePath."""

    def _node(index: int, node: CstNode) -> Node:
        return Node(node.tag_name, index, node.attributes)

    result = []
    children = tree
    for index in indexed_path:
        node = children[index]
        result.append(_node(index, node))
        children = node.children
    return result


def html_to_node_path(html: str, indexed_path: list[int]) -> NodePath:
    tree = html_to_tree(html)
    return tree_to_path(tree, indexed_path)



import shutil
import sys
from datetime import timedelta
from pathlib import Path
from time import sleep
from typing import List

from wwwpy.server.filesystem_sync import new_tmp_path, filesystemevents_print, Sync
from wwwpy.server.filesystem_sync import sync_zip
from wwwpy.common.filesystem.sync.event import Event
from wwwpy.server.filesystem_sync.watchdog_debouncer import WatchdogDebouncer


class WatchdogLocal:
    def __init__(self, tmp_path: Path, sync: Sync, exist_ok=False, print_changes=True):
        self.sync = sync
        self.source = tmp_path / 'source'
        self.source.mkdir(exist_ok=exist_ok)
        self.target = tmp_path / 'target'
        self.target.mkdir(exist_ok=exist_ok)
        self.window = timedelta(milliseconds=100)
        self.callback_count = 0

        def callback(events: List[Event]):
            assert events
            self.callback_count += 1
            filesystemevents_print(events)
            changes = self.sync.sync_source(self.source, events)
            self.sync.sync_target(self.target, changes)

        self.debounced_watcher = WatchdogDebouncer(self.source, self.window, callback)

    def start(self):
        self.debounced_watcher.start()

    def copy_source_to_target(self):
        shutil.rmtree(self.target, ignore_errors=True)
        shutil.copytree(self.source, self.target, dirs_exist_ok=True)

def main():
    tmp_path = Path(sys.argv[1]) if len(sys.argv) > 1 else new_tmp_path()

    local = WatchdogLocal(tmp_path, exist_ok=True, sync=sync_zip)
    local.copy_source_to_target()

    print(f'watching file://{local.source}')
    local.start()
    while True:
        sleep(10)


if __name__ == '__main__':
    main()



import logging
from typing import Callable

from js import console

from wwwpy.remote import set_timeout
from wwwpy.server import rpc


class _CustomHandler(logging.Handler):
    def __init__(self, callback: Callable[[str], None]):
        super().__init__()
        self._emit = callback

    def emit(self, record: logging.LogRecord):
        log_entry = self.format(record)
        self._emit(log_entry)


def redirect_logging():
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    for handler in root.handlers:
        if isinstance(handler, _CustomHandler):
            return

    def emit(msg: str):
        console.log(msg)

        async def _():
            await rpc.server_console_log(msg)

        set_timeout(_)

    formatter = logging.Formatter('%(asctime)s %(levelname).1s %(name)s - %(message)s')
    custom_handler = _CustomHandler(emit)
    custom_handler.setFormatter(formatter)
    root.addHandler(custom_handler)
